import pytest
from autobahn.wamp.types import RegisterOptions, SubscribeOptions

from asphalt.wamp.registry import WAMPRegistry


@pytest.fixture
def registry():
    return WAMPRegistry(procedure_defaults=RegisterOptions(concurrency=1),
                        subscription_defaults=SubscribeOptions(get_retained=True))


async def dummyhandler(ctx):
    pass


def test_defaults_as_dicts():
    registry = WAMPRegistry(procedure_defaults=dict(concurrency=1),
                            subscription_defaults=dict(get_retained=True))
    assert registry.procedure_defaults.concurrency == 1
    assert registry.subscription_defaults.get_retained


@pytest.mark.parametrize('use_dict', [False, True])
@pytest.mark.parametrize('use_decorator', [False, True])
def test_procedure(registry: WAMPRegistry, use_dict, use_decorator):
    options = dict(match='prefix', invoke='roundrobin')
    if not use_dict:
        options = RegisterOptions(**options)

    if use_decorator:
        registry.procedure('procedurename', options)(dummyhandler)
    else:
        registry.add_procedure(dummyhandler, 'procedurename', options)

    assert list(registry.procedures.keys()) == ['procedurename']
    procedure = registry.procedures['procedurename']
    assert procedure.handler is dummyhandler
    assert procedure.name == 'procedurename'
    assert procedure.options.concurrency == 1


def test_procedure_bad_argument_count(registry: WAMPRegistry):
    def invalid_procedure_handler():
        pass

    exc = pytest.raises(TypeError, registry.add_procedure, invalid_procedure_handler, 'procedure')
    exc.match('procedure handler must accept at least one positional argument')


def test_procedure_simple_decorator(registry: WAMPRegistry):
    registry.procedure(dummyhandler)
    assert list(registry.procedures.keys()) == ['dummyhandler']
    assert registry.procedures['dummyhandler'].handler is dummyhandler
    assert type(registry.procedures['dummyhandler'].options) is RegisterOptions


def test_duplicate_procedure(registry: WAMPRegistry):
    registry.add_procedure(dummyhandler, 'handler')

    exc = pytest.raises(ValueError, registry.add_procedure, dummyhandler, 'handler')
    exc.match('duplicate registration of procedure "handler"')


def test_subscriber_no_options(registry: WAMPRegistry):
    registry.add_subscriber(dummyhandler, 'topic')
    assert len(registry.subscriptions) == 1
    assert type(registry.subscriptions[0].options) is SubscribeOptions


@pytest.mark.parametrize('use_dict', [False, True])
@pytest.mark.parametrize('use_decorator', [False, True])
def test_subscriber(registry: WAMPRegistry, use_dict, use_decorator):
    options = dict(match='prefix')
    if not use_dict:
        options = SubscribeOptions(**options)

    if use_decorator:
        registry.subscriber('topic', options)(dummyhandler)
    else:
        registry.add_subscriber(dummyhandler, 'topic', options)

    assert len(registry.subscriptions) == 1
    subscription = registry.subscriptions[0]
    assert subscription.handler is dummyhandler
    assert subscription.topic == 'topic'
    assert subscription.options.match == 'prefix'
    assert subscription.options.get_retained


def test_subscriber_argument_count(registry: WAMPRegistry):
    def invalid_subscriber():
        pass

    exc = pytest.raises(TypeError, registry.add_subscriber, invalid_subscriber, 'topic')
    exc.match('subscriber must accept at least one positional argument')


@pytest.mark.parametrize('use_decorator', [False, True])
def test_exception_decorator(registry: WAMPRegistry, use_decorator):
    if use_decorator:
        registry.exception('x.y.z')(RuntimeError)
    else:
        registry.map_exception(RuntimeError, 'x.y.z')

    assert registry.exceptions == {'x.y.z': RuntimeError}


def test_invalid_exception(registry: WAMPRegistry):
    exc = pytest.raises(TypeError, registry.map_exception, str, 'x.y.z')
    exc.match('exc_type must be a subclass of BaseException')


@pytest.mark.parametrize('prefix', ['', 'middle'])
def test_add_from(prefix):
    child_registry = WAMPRegistry('child', procedure_defaults=RegisterOptions(invoke='roundrobin'),
                                  subscription_defaults=SubscribeOptions(match='prefix'))
    child_registry.add_procedure(dummyhandler, 'procedure', RegisterOptions(match='exact'))
    child_registry.add_subscriber(dummyhandler, 'topic', SubscribeOptions(match='exact'))
    child_registry.map_exception(RuntimeError, 'x.y.z')

    parent_registry = WAMPRegistry(
        'parent', procedure_defaults=RegisterOptions(match='prefix', invoke='first',
                                                     concurrency=2),
        subscription_defaults=SubscribeOptions(match='wildcard', get_retained=True)
    )
    parent_registry.add_from(child_registry, prefix)

    procedure_name = 'parent.{}child.procedure'.format((prefix + '.') if prefix else '')
    assert len(parent_registry.procedures) == 1
    assert parent_registry.procedures[procedure_name].options.match == 'exact'
    assert parent_registry.procedures[procedure_name].options.invoke == 'roundrobin'
    assert parent_registry.procedures[procedure_name].options.concurrency == 2

    expected_topic = 'parent.{}child.topic'.format((prefix + '.') if prefix else '')
    assert len(parent_registry.subscriptions) == 1
    assert parent_registry.subscriptions[0].topic == expected_topic
    assert parent_registry.subscriptions[0].options.match == 'exact'

    exception_name = 'parent.{}child.x.y.z'.format((prefix + '.') if prefix else '')
    assert parent_registry.exceptions == {exception_name: RuntimeError}


def test_repr(registry: WAMPRegistry):
    assert repr(registry) == "WAMPRegistry(prefix='')"



import ssl

from asphalt.serialization.api import Serializer
from asphalt.serialization.serializers.json import JSONSerializer
import pytest

from asphalt.wamp.client import WAMPClient
from asphalt.wamp.component import WAMPComponent
from asphalt.wamp.registry import WAMPRegistry


@pytest.mark.asyncio
async def test_single_client(context):
    tls_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    context.add_resource(tls_context)

    serializer = JSONSerializer()
    context.add_resource(serializer, types=[Serializer])

    registry = WAMPRegistry()
    context.add_resource(registry)

    component = WAMPComponent(tls_context='default', serializer='default', registry='default')
    await component.start(context)

    assert isinstance(context.wamp, WAMPClient)
    assert context.wamp.tls_context is tls_context
    assert context.wamp.serializer is serializer


@pytest.mark.asyncio
async def test_multiple_clients(context):
    component = WAMPComponent(clients={
        'wamp1': {'host': '192.168.10.1', 'port': 8085},
        'wamp2': {'path': '/'}
    }, auth_id='username')
    await component.start(context)

    assert isinstance(context.wamp1, WAMPClient)
    assert isinstance(context.wamp2, WAMPClient)
    assert context.wamp1.host == '192.168.10.1'
    assert context.wamp1.port == 8085
    assert context.wamp1.path == '/ws'
    assert context.wamp2.host == 'localhost'
    assert context.wamp2.port == 8080
    assert context.wamp2.path == '/'
    assert context.wamp1.auth_id == context.wamp2.auth_id == 'username'



"""This example demonstrates how to subscribe to topics with WAMP."""

import logging
import sys

from asphalt.core import ContainerComponent, Context, run_application
from asphalt.wamp.context import EventContext

logger = logging.getLogger(__name__)


def subscriber(ctx: EventContext, message: str):
    logger.info('Received message from %s: %s', ctx.topic, message)


class SubscriberComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp')
        await super().start(ctx)

        topic = sys.argv[1]
        await ctx.wamp.subscribe(subscriber, topic)
        logger.info('Subscribed to topic: %s', topic)


if len(sys.argv) < 2:
    print('Usage: {} <topic>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(SubscriberComponent(), logging=logging.INFO)



import inspect
from collections import OrderedDict
from typing import Callable, Dict, Any, Union, NamedTuple, TypeVar

from autobahn.wamp.types import SubscribeOptions, RegisterOptions
from typeguard import check_argument_types

__all__ = ('Procedure', 'Subscriber', 'WAMPRegistry')

Procedure = NamedTuple('Procedure', [('name', str), ('handler', Callable),
                                     ('options', RegisterOptions)])
Subscriber = NamedTuple('Subscriber', [('topic', str), ('handler', Callable),
                                       ('options', SubscribeOptions)])
T_Options = TypeVar('T_Options')


def _validate_handler(handler: Callable, kind: str) -> None:
    spec = inspect.getfullargspec(handler)
    min_args = 2 if inspect.ismethod(handler) else 1
    if not spec.varargs and len(spec.args) < min_args:
        raise TypeError('{} must accept at least one positional argument'.format(kind))


def _apply_defaults(options: T_Options, defaults: T_Options) -> None:
    for attrname in dir(options.__class__):
        if not attrname.startswith('_'):
            value = getattr(options, attrname)
            if value is None:
                default = getattr(defaults, attrname)
                setattr(options, attrname, default)


class WAMPRegistry:
    """
    Hosts a collection of WAMP procedure and subscriber registrations and exception mappings.

    The purpose of this class is to ease the task of collecting all of the procedure handlers,
    event listeners and exception mappings of the application into a single place that the client
    can then use to register those on the router when a session has been opened.
    The alternative would be to call :meth:`~asphalt.wamp.client.WAMPClient.register_procedure`,
    :meth:`~asphalt.wamp.client.WAMPClient.subscribe` and
    :meth:`~asphalt.wamp.client.WAMPClient.map_exception` after
    :meth:`~asphalt.wamp.client.WAMPClient.connect`. This would not be very modular, however, since
    the code the connects the client would have to know about every single handler callbacks in
    advance.

    :ivar procedures: registered procedure handlers
    :vartype procedures: Dict[str, Procedure]
    :ivar subscriptions: registered event subscribers
    :vartype subscriptions: List[Subscription]
    :ivar exceptions: mapping of WAMP error code to exception class
    :vartype exceptions: Dict[str, type]
    """

    __slots__ = ('prefix', 'procedure_defaults', 'subscription_defaults', 'procedures',
                 'subscriptions', 'exceptions')

    def __init__(self, prefix: str = '', *,
                 procedure_defaults: Union[RegisterOptions, Dict[str, Any]] = None,
                 subscription_defaults: Union[SubscribeOptions, Dict[str, Any]] = None):
        """
        :param prefix: a prefix that is added to the name of every registered procedure
        :param procedure_defaults: default values to use for omitted arguments to
            :meth:`add_procedure`
        :param subscription_defaults: default values to use for omitted arguments to
            :meth:`add_subscriber`

        """
        assert check_argument_types()
        if prefix and not prefix.endswith('.'):
            prefix += '.'

        self.prefix = prefix
        if isinstance(procedure_defaults, dict):
            procedure_defaults = RegisterOptions(**procedure_defaults)
        elif procedure_defaults is None:
            procedure_defaults = RegisterOptions()

        if isinstance(subscription_defaults, dict):
            subscription_defaults = SubscribeOptions(**subscription_defaults)
        elif subscription_defaults is None:
            subscription_defaults = SubscribeOptions()

        self.procedure_defaults = procedure_defaults
        self.subscription_defaults = subscription_defaults
        self.procedures = OrderedDict()
        self.subscriptions = []
        self.exceptions = OrderedDict()

    def add_procedure(self, handler: Callable, name: str = None,
                      options: Union[RegisterOptions, Dict[str, Any]] = None) -> Procedure:
        """
        Add a procedure handler.

        The callable will receive a :class:`~asphalt.wamp.context.CallContext` instance as its
        first argument. The other positional and keyword arguments will be the arguments the caller
        passed to it.

        :param handler: callable that handles the procedure calls
        :param name: name of the endpoint to register (relative to registry's prefix); if ``None``,
            use the callable's internal name
        :param options: either an Autobahn register options object or a dictionary of keyword
            arguments to make one
        :return: the procedure registration object
        :raises TypeError: if the handler does not accept at least a single positional argument
        :raises ValueError: if there is already a handler registered for this endpoint

        .. seealso:: `How registrations work <http://crossbar.io/docs/How-Registrations-Work/>`_
        .. seealso:: `Pattern based registrations
            <http://crossbar.io/docs/Pattern-Based-Registrations/>`_
        .. seealso:: `Shared registrations <http://crossbar.io/docs/Shared-Registrations/>`_

        """
        assert check_argument_types()
        _validate_handler(handler, 'procedure handler')

        if isinstance(options, dict):
            options = RegisterOptions(**options)
        elif options is None:
            options = RegisterOptions()

        _apply_defaults(options, self.procedure_defaults)
        name = self.prefix + (name or handler.__name__)
        registration = Procedure(name, handler, options)
        if self.procedures.setdefault(name, registration) is not registration:
            raise ValueError('duplicate registration of procedure "{}"'.format(name))

        return registration

    def procedure(self, name: Union[str, Callable] = None,
                  options: Union[RegisterOptions, Dict[str, Any]] = None) -> Callable:
        """
        Decorator version of :meth:`add_procedure`.

        Can be used as ``@procedure(...)`` or simply ``@procedure``.

        If ``name`` has not been specified, the function name of the handler is used.

        :param name: name of the endpoint to register (relative to registry's prefix)
        :param options: either an Autobahn register options object or a dictionary of keyword
            arguments to make one

        """
        def wrapper(handler: Callable):
            self.add_procedure(handler, name, options)
            return handler

        if inspect.isfunction(name):
            handler, name = name, None
            return wrapper(handler)

        return wrapper

    def add_subscriber(self, handler: Callable, topic: str,
                       options: Union[SubscribeOptions, Dict[str, Any]] = None) -> Subscriber:
        """
        Decorator that registers a callable to receive events on the given topic.

        The callable  will receive an :class:`~asphalt.wamp.context.EventContext` instance as its
        first argument and must return an awaitable.
        The other positional and keyword arguments will be the arguments the publisher passed to
        the ``publish()`` method.

        The topic may contain wildcards (``a..b``).

        :param handler: callable that receives matching events
        :param topic: the topic to listen to
        :param options: either an Autobahn subscribe options object or a dictionary of keyword
            arguments to make one
        :return: the subscription registration object
        :raises TypeError: if the handler is not a coroutine or does not accept at least a single
            positional argument

        .. seealso:: `How subscriptions work <http://crossbar.io/docs/How-Subscriptions-Work/>`_
        .. seealso:: `Pattern based subscriptions
            <http://crossbar.io/docs/Pattern-Based-Subscriptions/>`_

        """
        assert check_argument_types()
        _validate_handler(handler, 'subscriber')

        if isinstance(options, dict):
            options = SubscribeOptions(**options)
        elif options is None:
            options = SubscribeOptions()

        _apply_defaults(options, self.subscription_defaults)
        topic = self.prefix + (topic or handler.__name__)
        subscription = Subscriber(topic, handler, options)
        self.subscriptions.append(subscription)
        return subscription

    def subscriber(self, topic: str,
                   options: Union[SubscribeOptions, Dict[str, Any]] = None) -> Callable:
        """
        Decorator version of :meth:`add_subscriber`.

        :param topic: the topic to listen to
        :param options: either an Autobahn subscribe options object or a dictionary of keyword
            arguments to make one

        .. seealso:: `How subscriptions work <http://crossbar.io/docs/How-Subscriptions-Work/>`_
        .. seealso:: `Pattern based subscriptions
            <http://crossbar.io/docs/Pattern-Based-Subscriptions/>`_

        """
        def wrapper(handler: Callable):
            self.add_subscriber(handler, topic, options)
            return handler

        return wrapper

    def map_exception(self, exc_type: type, code: str) -> None:
        """
        Map a Python exception class to a WAMP error code.

        :param code: the WAMP error code (e.g. ``foo.bar.baz``)
        :param exc_type: an exception class
        :raises TypeError: if ``exc_type`` does not inherit from :class`BaseException`

        """
        assert check_argument_types()
        if not isinstance(exc_type, type) or not issubclass(exc_type, BaseException):
            raise TypeError('exc_type must be a subclass of BaseException')

        self.exceptions[self.prefix + code] = exc_type

    def exception(self, error: str) -> Callable:
        """
        Decorator version of :meth:`map_exception`.

        :param error: the WAMP error code (e.g. ``foo.bar.baz``)

        """
        def wrapper(exc_class: type):
            self.map_exception(exc_class, error)
            return exc_class

        return wrapper

    def add_from(self, registry: 'WAMPRegistry', prefix: str = '') -> None:
        """
        Add all the procedures, subscribers and exception mappings from another registry to this
        one.

        If no prefix has been specified, the final name of each added procedure/subscriber endpoint
        or mapped error will be of the form ``<root prefix>.<name>``.
        If a prefix has been specified, the name will be ``<root prefix>.<prefix>.<name>``.

        :param registry: a WAMP registry
        :param prefix: prefix to add to names of all procedure endpoints

        """
        assert check_argument_types()
        if prefix and not prefix.endswith('.'):
            prefix += '.'

        for registration in registry.procedures.values():
            self.add_procedure(registration.handler, prefix + registration.name,
                               registration.options)

        for subscription in registry.subscriptions:
            self.add_subscriber(subscription.handler, prefix + subscription.topic,
                                subscription.options)

        for error, exc_type in registry.exceptions.items():
            self.map_exception(exc_type, prefix + error)

    def __repr__(self, *args, **kwargs):
        return '{}(prefix={!r})'.format(self.__class__.__name__, self.prefix)



"""
This example demonstrates the use of the progress reporting feature in RPC calls.
It (ab)uses the feature to send files to the clients in chunks.

The command line argument should be a path to a directory where files will be served from.
"""

from pathlib import Path
import logging
import sys

from asphalt.core import ContainerComponent, Context, run_application
from asphalt.serialization.serializers.cbor import CBORSerializer
from asphalt.wamp.context import CallContext
from async_generator import aclosing
from asyncio_extras.file import open_async

logger = logging.getLogger(__name__)


async def send_file(ctx: CallContext, path: str):
    final_path = ctx.base_path / path
    if not final_path.is_file():
        raise Exception('{} is not a file'.format(path))

    async with open_async(final_path, 'rb') as f:
        logger.info('Sending %s', path)
        async with aclosing(f.async_readchunks(65536)) as stream:
            async for chunk in stream:
                ctx.progress(chunk)

    logger.info('Finished sending %s', path)


class FileServerComponent(ContainerComponent):
    async def start(self, ctx: Context):
        # Need either msgpack or CBOR to serialize binary messages (such as raw file chunks)
        self.add_component('wamp', serializer=CBORSerializer())
        await super().start(ctx)

        ctx.base_path = Path(sys.argv[1])
        await ctx.wamp.register(send_file, 'send_file')


if len(sys.argv) < 2:
    print('Usage: {} <base directory>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(FileServerComponent(), logging=logging.INFO)



"""This example demonstrates how to make RPC calls with WAMP."""

import logging
import sys

from asphalt.core import CLIApplicationComponent, Context, run_application

logger = logging.getLogger(__name__)


class RPCClientComponent(CLIApplicationComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp')
        await super().start(ctx)

    async def run(self, ctx: Context):
        message = sys.argv[1]
        result = await ctx.wamp.call('uppercase', message)
        logger.info('%r translated to upper case is %r', message, result)


if len(sys.argv) < 2:
    print('Usage: {} <text>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(RPCClientComponent(), logging=logging.INFO)



import os

import pytest
import txaio
from async_generator import async_generator, yield_

from asphalt.core import Context
from asphalt.wamp.client import WAMPClient


@pytest.fixture(autouse=True)
def setup_txaio(event_loop):
    txaio.use_asyncio()
    txaio.config.loop = event_loop


@pytest.fixture
@async_generator
async def context():
    async with Context() as ctx:
        await yield_(ctx)


@pytest.fixture
@async_generator
async def wampclient(request, context):
    kwargs = getattr(request, 'param', {})
    kwargs.setdefault('host', os.getenv('CROSSBAR_HOST', '127.0.0.1'))
    kwargs.setdefault('max_reconnection_attempts', 0)
    client = WAMPClient(**kwargs)
    await client.start(context)
    await yield_(client)
    await client.stop()



"""This example demonstrates how to publish messages with WAMP."""

import logging
import sys

from asphalt.core import CLIApplicationComponent, Context, run_application

logger = logging.getLogger(__name__)


class PublisherComponent(CLIApplicationComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp')
        await super().start(ctx)

    async def run(self, ctx: Context):
        topic, message = sys.argv[1:3]
        await ctx.wamp.publish(topic, message)


if len(sys.argv) < 3:
    print('Usage: {} <topic> <message>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(PublisherComponent(), logging=logging.INFO)



from autobahn.wamp import serializer as autobahn
from asphalt.serialization.serializers.cbor import CBORSerializer
from asphalt.serialization.serializers.json import JSONSerializer
from asphalt.serialization.serializers.msgpack import MsgpackSerializer
from asphalt.serialization.serializers.yaml import YAMLSerializer
import pytest

from asphalt.wamp.serializers import wrap_serializer, ObjectSerializerWrapper


@pytest.fixture(params=[JSONSerializer, MsgpackSerializer, CBORSerializer],
                ids=['json', 'msgpack', 'cbor'])
def serializer(request):
    return request.param()


@pytest.mark.parametrize('value', [
    'text', 123, ['a', 1, 2.5, True], {'x': 'a', 'y': 2.5, 'z': True}
], ids=['string', 'int', 'list', 'dict'])
def test_object_serializer_wrapper(serializer, value):
    binary = serializer.mimetype != 'application/json'
    wrapper = ObjectSerializerWrapper(serializer, binary)
    payload = wrapper.serialize(value)
    objects = wrapper.unserialize(payload)
    assert len(objects) == 1
    assert objects[0] == value


def test_wrap_serializer(serializer):
    wrapped = wrap_serializer(serializer)
    assert isinstance(wrapped, autobahn.Serializer)


def test_wrap_unsupported_serializer():
    exc = pytest.raises(TypeError, wrap_serializer, YAMLSerializer())
    exc.match('cannot adapt serializer of type text/yaml')



from autobahn.wamp.types import SessionDetails, CloseDetails

from asphalt.core import Event

__all__ = ('SessionJoinEvent', 'SessionLeaveEvent')


class SessionJoinEvent(Event):
    """
    Signals that the client has joined the WAMP realm on the router.

    :ivar details: the autobahn-provided session details
    :vartype details: ~autobahn.wamp.types.SessionDetails
    """

    __slots__ = 'details'

    def __init__(self, source, topic: str, session_details: SessionDetails):
        super().__init__(source, topic)
        self.details = session_details


class SessionLeaveEvent(Event):
    """
    Signals that the client has left the WAMP realm on the router.

    :ivar str reason: the reason why the client left the realm
    :ivar str message: the closing message
    """

    __slots__ = 'reason', 'message'

    def __init__(self, source, topic: str, close_details: CloseDetails):
        super().__init__(source, topic)
        self.reason = close_details.reason
        self.message = close_details.message



"""
This example demonstrates the use of the progress reporting feature in RPC calls.
It (ab)uses the feature to download a file from the server (server.py) in chunks to a temporary
directory.

The file path parameter should be a path relative to the directory where the server is serving
files from.
"""

import logging
import sys
from pathlib import Path
from tempfile import mkdtemp

from asphalt.core import CLIApplicationComponent, Context, run_application
from asphalt.serialization.serializers.cbor import CBORSerializer

logger = logging.getLogger(__name__)


class FileGetterComponent(CLIApplicationComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp', serializer=CBORSerializer())
        await super().start(ctx)

    async def run(self, ctx: Context):
        def on_progress(data: bytes):
            # This gets called for every chunk the server sends (ctx.progress() on the other side)
            outfile.write(data)
            print('\r{} bytes written'.format(outfile.tell()), end='')

        remote_path = sys.argv[1]
        local_path = Path(mkdtemp()) / Path(remote_path).name
        with local_path.open('wb') as outfile:
            await ctx.wamp.call('send_file', remote_path, on_progress=on_progress)

        print('\nFile saved as %s' % local_path)


if len(sys.argv) < 2:
    print('Usage: {} <file path>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(FileGetterComponent(), logging=logging.INFO)



from autobahn.wamp.types import SessionDetails, CloseDetails

from asphalt.wamp.events import SessionJoinEvent, SessionLeaveEvent


def test_join_event(wampclient):
    details = SessionDetails('default', 5)
    event = SessionJoinEvent(wampclient, 'realm_joined', details)
    assert event.details.session == 5


def test_leave_event(wampclient):
    details = CloseDetails('reason', 'message')
    event = SessionLeaveEvent(wampclient, 'realm_left', details)
    assert event.reason == 'reason'
    assert event.message == 'message'



from autobahn.wamp.types import CallDetails, SessionDetails, EventDetails

from asphalt.core import Context

__all__ = ('CallContext', 'EventContext')


class CallContext(Context):
    """
    Context class for procedure calls.

    Procedure call handlers are passed an instance of this class as the first argument.

    :ivar int session_id: our own WAMP session ID
    :ivar Optional[Callable] progress: a callable through which the handler can send
        progress information to the caller
    :ivar Optional[int] caller_session_id: WAMP session ID of the caller (if disclosed)
    :ivar Optional[str] caller_auth_id: WAMP authentication ID (username) of the caller
        (if disclosed)
    :ivar Optional[str] caller_auth_role: WAMP authentication role of the caller (if disclosed)
    :ivar Optional[str] procedure: the actual name of the procedure (when using a pattern based
        registration)
    :ivar Optional[str] enc_algo: payload encryption algorithm that was in use, if any
        (e.g. `cryptobox`, or a custom algorithm)
    """

    def __init__(self, parent: Context, session_details: SessionDetails,
                 call_details: CallDetails, **kwargs):
        super().__init__(parent, **kwargs)
        self.session_id = session_details.session
        self.progress = call_details.progress
        self.caller_session_id = call_details.caller
        self.caller_auth_id = call_details.caller_authid
        self.caller_auth_role = call_details.caller_authrole
        self.procedure = call_details.procedure
        self.enc_algo = call_details.enc_algo


class EventContext(Context):
    """
    Context class for WAMP events.

    Event subscribers are passed an instance of this class as the first argument.

    :ivar int session_id: our own WAMP session ID
    :ivar int publication_id: publication ID of the event
    :ivar int publisher_session_id: WAMP session ID of the publisher, if disclosed
    :ivar Optional[str] publisher_auth_id: WAMP authentication ID (username) of the publisher
        (if disclosed)
    :ivar Optional[str] publisher_auth_role: WAMP authentication role of the publisher
        (if disclosed)
    :ivar str topic: the exact topic the event was received on
    :ivar Optional[str] enc_algo: payload encryption algorithm that was in use, if any
        (e.g. `cryptobox`, or a custom algorithm)
    """

    def __init__(self, parent: Context, session_details: SessionDetails,
                 event_details: EventDetails, **kwargs):
        super().__init__(parent, **kwargs)
        self.session_id = session_details.session
        self.publication_id = event_details.publication
        self.publisher_session_id = event_details.publisher
        self.publisher_auth_id = event_details.publisher_authid
        self.publisher_auth_role = event_details.publisher_authrole
        self.topic = event_details.topic
        self.enc_algo = event_details.enc_algo



from autobahn.wamp.request import Registration, Subscription
from autobahn.wamp.types import SessionDetails, CallDetails, EventDetails
from asphalt.core.context import Context

from asphalt.wamp.context import CallContext, EventContext


def test_call_context():
    def progress(arg):
        pass

    parent = Context()
    session_details = SessionDetails('default', 5)
    registration = Registration(5420493024, 12313423, lambda: None, 'foo.bar')
    call_details = CallDetails(registration, progress=progress, caller=8, caller_authid='user',
                               caller_authrole='role', procedure='procedurename')
    context = CallContext(parent, session_details, call_details)

    assert context.session_id == 5
    assert context.caller_session_id == 8
    assert context.caller_auth_id == 'user'
    assert context.caller_auth_role == 'role'
    assert context.procedure == 'procedurename'
    assert context.enc_algo is None
    assert context.progress is progress


def test_event_context():
    parent = Context()
    session_details = SessionDetails('default', 5)
    subscription = Subscription(1230643, 'foo.bar', 5420493024, lambda: None)
    event_details = EventDetails(subscription, publication=15, publisher=8,
                                 publisher_authid='user', publisher_authrole='role', topic='topic')
    context = EventContext(parent, session_details, event_details)

    assert context.session_id == 5
    assert context.publisher_session_id == 8
    assert context.publisher_auth_id == 'user'
    assert context.publisher_auth_role == 'role'
    assert context.publication_id == 15
    assert context.topic == 'topic'
    assert context.enc_algo is None



"""This example demonstrates how to register a remote procedure handler with WAMP."""

import logging

from asphalt.core import ContainerComponent, Context, run_application


def uppercase(ctx, message: str):
    return message.upper()


class RPCServerComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp')
        await super().start(ctx)
        await ctx.wamp.register(uppercase, 'uppercase')


run_application(RPCServerComponent(), logging=logging.INFO)



"""
This example demonstrates how to authenticate against the WAMP router with the "ticket" method.
It will make a call to the server that returns with a message containing the authentication ID
of this client.
"""

import logging

from asphalt.core import CLIApplicationComponent, Context, run_application

logger = logging.getLogger(__name__)


class RPCClientComponent(CLIApplicationComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp', auth_method='ticket',
                           auth_id='testclient', auth_secret='client123')
        await super().start(ctx)

    async def run(self, ctx: Context):
        result = await ctx.wamp.call('hello')
        logger.info('Server responded: %s', result)


run_application(RPCClientComponent(), logging=logging.INFO)



"""
This example WAMP client first calls the ``test.reverse`` remote method with string given as the
command line. It then publishes the return value on the ``topics.subtopic`` topic.
"""

import logging
import sys

from asphalt.core import CLIApplicationComponent, Context, run_application


class RPCClientComponent(CLIApplicationComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp')
        await super().start(ctx)

    async def run(self, ctx: Context):
        message = sys.argv[1]
        result = await ctx.wamp.call('test.reverse', message)
        await ctx.wamp.publish('topics.subtopic', result)


if len(sys.argv) < 2:
    print('Usage: {} <text>'.format(sys.argv[0]), file=sys.stderr)
    sys.exit(1)

run_application(RPCClientComponent(), logging=logging.INFO)



import asyncio
import logging
from asyncio import (  # noqa: F401
    wait, sleep, Future, Task, shield, AbstractEventLoop, CancelledError, gather)
from contextlib import suppress
from inspect import isawaitable
from ssl import SSLContext
from typing import Callable, Optional, Union, Set, Dict, Any  # noqa: F401
from warnings import warn

from asphalt.core import Context, resolve_reference, Signal
from asphalt.exceptions import report_exception
from asphalt.serialization.api import Serializer
from asphalt.serialization.serializers.json import JSONSerializer
from async_timeout import timeout
from autobahn.asyncio.wamp import ApplicationSession
from autobahn.asyncio.websocket import WampWebSocketClientFactory
from autobahn.wamp import auth, ApplicationError, SessionNotReady, TransportLost
from autobahn.wamp.types import (
    ComponentConfig, SessionDetails, EventDetails, CallDetails, PublishOptions, CallOptions,
    Challenge, SubscribeOptions, RegisterOptions)
from typeguard import check_argument_types

from asphalt.wamp.context import CallContext, EventContext
from asphalt.wamp.events import SessionJoinEvent, SessionLeaveEvent
from asphalt.wamp.registry import WAMPRegistry, Procedure, Subscriber
from asphalt.wamp.serializers import wrap_serializer

__all__ = ('ConnectionError', 'WAMPClient')

logger = logging.getLogger(__name__)


class ConnectionError(Exception):
    """Raised when there was an error connecting to the WAMP router."""


class AsphaltSession(ApplicationSession):
    def __init__(self, realm: str, auth_method: Optional[str], auth_id: Optional[str],
                 auth_secret: Optional[str]) -> None:
        super().__init__(ComponentConfig(realm))
        self.__auth_method = auth_method
        self.__auth_id = auth_id
        self.__auth_secret = auth_secret

    def onConnect(self):
        self.join(self.config.realm, [self.__auth_method], self.__auth_id)

    def onChallenge(self, challenge: Challenge):
        if challenge.method != self.__auth_method:
            raise ConnectionError(
                'expected authentication method "{}" but received a "{}" challenge instead'.
                format(self.__auth_method, challenge.method))

        if challenge.method == 'wampcra':
            key = self.__auth_secret
            if 'salt' in challenge.extra:
                # salted secret
                key = auth.derive_key(self.__auth_secret.encode('utf-8'),
                                      challenge.extra['salt'], challenge.extra['iterations'],
                                      challenge.extra['keylen'])

            return auth.compute_wcs(key, challenge.extra['challenge'])
        elif challenge.method == 'ticket':  # ticket
            return self.__auth_secret


class WAMPClient:
    """
    A WAMP client.

    :ivar Signal realm_joined: a signal (:class:`~asphalt.wamp.events.SessionJoinEvent`) dispatched
        when the client has joined the realm and has registered any procedures and subscribers on
        the router
    :ivar Signal realm_left: a signal (:class:`~asphalt.wamp.events.SessionLeaveEvent`) dispatched
        when the client has left the realm
    """

    realm_joined = Signal(SessionJoinEvent)
    realm_left = Signal(SessionLeaveEvent)

    def __init__(self, host: str = 'localhost', port: int = 8080, path: str = '/ws',
                 realm: str = 'realm1', *, protocol_options: Dict[str, Any] = None,
                 connection_timeout: float = 10, reconnect_delay: float = 5,
                 max_reconnection_attempts: Optional[int] = 15,
                 shutdown_timeout: Optional[float] = 15,
                 registry: Union[WAMPRegistry, str] = None, tls: bool = False,
                 tls_context: Union[str, SSLContext] = None,
                 serializer: Union[Serializer, str] = None, auth_method: str = 'anonymous',
                 auth_id: str = None, auth_secret: str = None):
        """
        :param host: host address of the WAMP router
        :param port: port to connect to
        :param path: HTTP path on the router
        :param realm: the WAMP realm to join the application session to (defaults to the resource
            name if not specified)
        :param protocol_options: dictionary of Autobahn's `websocket protocol options`_
        :param connection_timeout: maximum time to wait for the client to connect to the router and
            join a realm
        :param reconnect_delay: delay between connection attempts (in seconds)
        :param max_reconnection_attempts: maximum number of connection attempts before giving up
        :param shutdown_timeout: maximum number of seconds to wait for the client to complete its
            shutdown sequence (unregister procedures/subscriptions, wait for running handlers to
            finish, leave the realm)
        :param registry: a :class:`~asphalt.wamp.registry.WAMPRegistry` instance, a
            ``module:varname`` reference or resource name of one
        :param tls: ``True`` to use TLS when connecting to the router
        :param tls_context: an :class:`~ssl.SSLContext` instance or the resource name of one
        :param serializer: a :class:`asphalt.serialization.api.Serializer` instance or the resource
            name of one
        :param auth_method: authentication method to use (valid values are currently ``anonymous``,
            ``wampcra`` and ``ticket``)
        :param auth_id: authentication ID (username)
        :param auth_secret: secret to use for authentication (ticket or password)

        .. _websocket protocol options:
            http://autobahn.readthedocs.io/en/latest/websocket/programming.html#websocket-options

        """
        assert check_argument_types()
        self.host = host
        self.port = port
        self.path = path
        self.reconnect_delay = reconnect_delay
        self.connection_timeout = connection_timeout
        self.shutdown_timeout = shutdown_timeout
        self.max_reconnection_attempts = max_reconnection_attempts
        self.realm = realm
        self.protocol_options = protocol_options or {}
        self.tls = tls
        self.tls_context = tls_context
        self.serializer = serializer or JSONSerializer()
        self.auth_method = auth_method
        self.auth_id = auth_id
        self.auth_secret = auth_secret

        self._parent_context = None  # type: Context
        self._loop = None  # type: AbstractEventLoop
        self._registry = resolve_reference(registry) or WAMPRegistry()
        self._session = None  # type: AsphaltSession
        self._session_details = None  # type: SessionDetails
        self._connect_task = None  # type: Task
        self._request_tasks = set()  # type: Set[Task]
        self._registrations = []
        self._subscriptions = []

    async def start(self, ctx: Context):
        self._parent_context = ctx
        self._loop = ctx.loop

        if isinstance(self.tls_context, str):
            self.tls_context = await ctx.request_resource(SSLContext, self.tls_context)
        if isinstance(self.serializer, str):
            self.serializer = await ctx.request_resource(Serializer, self.serializer)
        if isinstance(self._registry, str):
            self._registry = await ctx.request_resource(WAMPRegistry, self._registry)

        registry = WAMPRegistry()
        if self._registry:
            registry.add_from(self._registry)

        self._registry = registry

    async def close(self):
        warn('close() has been renamed to stop()', DeprecationWarning)
        await self.stop()

    async def stop(self):
        """
        Finish outstanding tasks and then disconnect from the router.

        First, all subscriptions and registrations are undone to prevent more publications or calls
        from coming in. Next, all outstanding tasks are awaited on. Finally, the client leaves the
        realm and disconnects from the router.

        """
        if self._session:
            sub_futures = [sub.unsubscribe() for sub in self._subscriptions if sub.active]
            proc_futures = [reg.unregister() for reg in self._registrations if reg.active]
            try:
                with timeout(self.shutdown_timeout):
                    if sub_futures or proc_futures:
                        logger.info(
                            'Unsubscribing %d subscriptions and unregistering %d procedures',
                            len(sub_futures), len(proc_futures))
                        await wait(sub_futures + proc_futures)

                    if self._request_tasks:
                        logger.info(
                            'Waiting for %d WAMP subscription/procedure handler tasks to finish',
                            len(self._request_tasks))
                        await wait(self._request_tasks)

                    with suppress(SessionNotReady):
                        await self._session.leave()
            except TransportLost:
                pass
            except asyncio.TimeoutError:
                self._session.disconnect()
        elif self._connect_task and not self._connect_task.done():
            self._connect_task.cancel()
            await gather(self._connect_task, return_exceptions=True)
            self._connect_task = None

    def map_exception(self, exc_class: type, error: str) -> None:
        """
        Map a Python exception to a WAMP error.

        :param exc_class: an exception class
        :param error: the WAMP error code

        """
        self._registry.map_exception(exc_class, error)
        if self._session:
            self._session.define(exc_class, error)

    async def _register(self, procedure: Procedure):
        async def wrapper(*args, _call_details: CallDetails, **kwargs):
            current_task = Task.current_task(loop=self._loop)
            self._request_tasks.add(current_task)
            async with CallContext(
                    self._parent_context, self._session_details, _call_details) as ctx:
                try:
                    retval = procedure.handler(ctx, *args, **kwargs)
                    if isawaitable(retval):
                        retval = await retval
                except ApplicationError:
                    raise  # These are deliberately raised so no need to report them
                except Exception:
                    report_exception(
                        ctx, 'Error running handler for procedure {!r}'.format(procedure.name),
                        logger=False)
                    raise
                finally:
                    self._request_tasks.remove(current_task)

                return retval

        procedure.options.details_arg = '_call_details'
        registration = await self._session.register(wrapper, procedure.name, procedure.options)
        self._registrations.append(registration)

    async def _subscribe(self, subscriber: Subscriber):
        async def wrapper(*args, details: EventDetails, **kwargs):
            current_task = Task.current_task(loop=self._loop)
            self._request_tasks.add(current_task)
            async with EventContext(self._parent_context, self._session_details,
                                    details) as ctx:
                try:
                    retval = subscriber.handler(ctx, *args, **kwargs)
                    if isawaitable(retval):
                        await retval
                except Exception:
                    report_exception(
                        ctx, 'Error running subscription handler for topic {!r}'.format(
                            subscriber.topic),
                        logger=False)
                    raise
                finally:
                    self._request_tasks.remove(current_task)

        subscriber.options.details = True
        subscriber.options.details_arg = 'details'
        subscription = await self._session.subscribe(wrapper, subscriber.topic, subscriber.options)
        self._subscriptions.append(subscription)

    async def register(self, handler: Callable, name: str = None,
                       options: Union[RegisterOptions, Dict[str, Any]] = None) -> None:
        """
        Add a procedure handler to the registry and attempt to register it on the router.

        :param handler: callable that handles calls for the given endpoint
        :param name: name of the endpoint to register (e.g. ``x.y.z``); omit to use the internal
            name of the callable
        :param options: either an Autobahn register options object or a dictionary of keyword
            arguments to make one

        .. note:: the ``details_arg`` option is set by WAMPClient itself so do not attempt to set
                  it yourself.

        """
        assert check_argument_types()
        if isinstance(options, dict):
            options = RegisterOptions(**options)
        elif options is None:
            options = RegisterOptions()

        procedure = self._registry.add_procedure(handler, name, options)
        if self._session is None:
            await self.connect()  # this will automatically register the procedure
        else:
            await self._register(procedure)

    async def subscribe(self, handler: Callable, topic: str,
                        options: Union[SubscribeOptions, Dict[str, Any]] = None) -> None:
        """
        Add a WAMP event subscriber to the registry and attempt to register it on the router.

        :param handler: the callable that is called when a message arrives
        :param topic: topic to subscribe to
        :param options: either an Autobahn subscribe options object or a dictionary of keyword
            arguments to make one

        .. note:: the ``details`` option is set by WAMPClient itself so do not attempt to set it
                  yourself.

        """
        assert check_argument_types()
        if isinstance(options, dict):
            options = SubscribeOptions(**options)
        elif options is None:
            options = SubscribeOptions()

        subscriber = self._registry.add_subscriber(handler, topic, options)
        if self._session is None:
            await self.connect()  # this will automatically register the subscriber
        else:
            await self._subscribe(subscriber)

    async def publish(self, topic: str, *args,
                      options: Union[PublishOptions, Dict[str, Any]] = None,
                      **kwargs) -> Optional[int]:
        """
        Publish an event on the given topic.

        :param topic: the topic to publish on
        :param args: positional arguments to pass to subscribers
        :param options: either an Autobahn publish options object or a dictionary of keyword
            arguments to make one
        :param kwargs: keyword arguments to pass to subscribers
        :return: publication ID (if the ``acknowledge`` option is ``True``)

        """
        assert check_argument_types()
        if self._session is None:
            await self.connect()

        if isinstance(options, dict):
            options = PublishOptions(**options)

        retval = self._session.publish(topic, *args, options=options, **kwargs)
        if options and options.acknowledge:
            publication = await retval
            return publication.id

    async def call(self, endpoint: str, *args, options: Union[CallOptions, Dict[str, Any]] = None,
                   **kwargs):
        """
        Call an RPC function.

        :param endpoint: name of the endpoint to call
        :param args: positional arguments to call the endpoint with
        :param options: either an Autobahn call options object or a dictionary of keyword arguments
            to make one
        :param kwargs: keyword arguments to call the endpoint with
        :return: the return value of the call
        :raises TimeoutError: if the call times out

        """
        assert check_argument_types()
        if self._session is None:
            await self.connect()

        if isinstance(options, dict):
            options = CallOptions(**options)

        return await self._session.call(endpoint, *args, options=options, **kwargs)

    def connect(self) -> Future:
        """
        Connect to the WAMP router and join the designated realm.

        When the realm is successfully joined, exceptions, procedures and event subscriptions from
        the registry are automatically registered with the router.

        The connection process is restarted if connection, joining the realm or registering the
        exceptions/procedures/subscriptions fails. If ``max_connection_attempts`` is set, it will
        limit the number of attempts. If this limit is reached, the future gets the last
        exception set to it. Otherwise, the process is repeated indefinitely until it succeeds.

        If the realm has already been joined, the future completes instantly.

        :raises ConnectionError: if there is a protocol level problem connecting to the router

        """
        async def do_connect() -> None:
            def create_session() -> AsphaltSession:
                session = AsphaltSession(self.realm, self.auth_method, self.auth_id,
                                         self.auth_secret)
                session.on('disconnect', on_disconnect)
                session.on('join', on_join)
                session.on('leave', on_leave)
                return session

            def on_disconnect(session: AsphaltSession, was_clean: bool):
                if not was_clean:
                    join_future.set_exception(ConnectionError('connection closed unexpectedly'))

            def on_join(session: AsphaltSession, details: SessionDetails):
                session.off('disconnect')
                join_future.set_result((session, details))

            def on_leave(session: AsphaltSession, details):
                self._session = None
                self._session_details = None
                self._connect_task = None
                self._subscriptions.clear()
                self._registrations.clear()
                self.realm_left.dispatch(details)
                if not session._goodbye_sent:
                    if not join_future.done():
                        join_future.set_exception(ConnectionError(details.message))
                    elif join_future.done():
                        logger.error('Connection lost; reconnecting')
                        self.connect()

            proto = 'wss' if self.tls else 'ws'
            url = '{proto}://{self.host}:{self.port}{self.path}'.format(proto=proto, self=self)
            logger.info('Connecting to %s', url)
            serializers = [wrap_serializer(self.serializer)]
            attempts = 0

            while self._session is None:
                transport = None
                attempts += 1
                try:
                    join_future = self._loop.create_future()
                    transport_factory = WampWebSocketClientFactory(
                        create_session, url=url, serializers=serializers, loop=self._loop)
                    transport_factory.setProtocolOptions(**self.protocol_options)
                    with timeout(self.connection_timeout):
                        transport, protocol = await self._loop.create_connection(
                            transport_factory, self.host, self.port,
                            ssl=self.tls_context or True if self.tls else False)

                        # Connection established; wait for the session to join the realm
                        logger.info('Connected to %s; attempting to join realm %s', self.host,
                                    self.realm)
                        self._session, self._session_details = await join_future

                    # Register exception mappings with the session
                    logger.info(
                        'Realm %r joined; registering %d procedure(s), %d subscription(s) and %d '
                        'exception(s)', self._session_details.realm,
                        len(self._registry.procedures), len(self._registry.subscriptions),
                        len(self._registry.exceptions))
                    for error, exc_type in self._registry.exceptions.items():
                        self._session.define(exc_type, error)

                    # Register procedures with the session
                    for procedure in self._registry.procedures.values():
                        with timeout(10):
                            await self._register(procedure)

                    # Register subscribers with the session
                    for subscriber in self._registry.subscriptions:
                        with timeout(10):
                            await self._subscribe(subscriber)
                except Exception as e:
                    if self._session:
                        await self.stop()
                    elif transport:
                        transport.close()

                    if isinstance(e, CancelledError):
                        logger.info('Connection attempt cancelled')
                        raise

                    if (self.max_reconnection_attempts is not None and
                            attempts > self.max_reconnection_attempts):
                        raise

                    logger.warning('Connection failed (attempt %d): %s(%s); reconnecting in %s '
                                   'seconds', attempts, e.__class__.__name__, e,
                                   self.reconnect_delay)
                    await sleep(self.reconnect_delay)

            # Notify listeners that we've joined the realm
            self.realm_joined.dispatch(self._session_details)
            logger.info('Joined realm %r', self.realm)

        # Start a new connection attempt only if not connected and there is no attempt in progress
        if not self._connect_task or self._connect_task.cancelled():
            self._connect_task = self._loop.create_task(do_connect())
            return self._connect_task
        else:
            return self._connect_task if self._connect_task.done() else shield(self._connect_task)

    @property
    def session_id(self) -> Optional[int]:
        """
        Return the current WAMP session ID.

        :return: the session ID or ``None`` if not in a session.

        """
        return self._session_details.session if self._session_details else None

    @property
    def details(self) -> Optional[SessionDetails]:
        """
        Return the session details object provided by Autobahn if the session has been established.

        """
        return self._session_details



import asyncio
import logging
import os
import re
from typing import Dict, Any

import pytest
from autobahn.wamp import ApplicationError
from autobahn.wamp.types import Challenge, PublishOptions, CallOptions

from asphalt.core import executor, Context, qualified_name
from asphalt.exceptions import ExtrasProvider
from asphalt.exceptions.api import ExceptionReporter
from asphalt.wamp.client import WAMPClient, AsphaltSession, ConnectionError
from asphalt.wamp.events import SessionJoinEvent, SessionLeaveEvent
from asphalt.wamp.extras_providers import WAMPExtrasProvider


class TestAsphaltSession:
    @pytest.fixture
    def session(self, request):
        return AsphaltSession('default', request.param, 'foo', 'bar')

    @pytest.mark.parametrize('session', ['ticket'], indirect=['session'])
    def test_challenge_mismatch(self, session):
        challenge = Challenge('wampcra')
        exc = pytest.raises(ConnectionError, session.onChallenge, challenge)
        assert exc.match('expected authentication method "ticket" but received a "wampcra" '
                         'challenge instead')

    @pytest.mark.parametrize('session', ['ticket'], indirect=['session'])
    def test_ticket_challenge(self, session):
        challenge = Challenge('ticket')
        assert session.onChallenge(challenge) == 'bar'

    @pytest.mark.parametrize('session', ['wampcra'], indirect=['session'])
    def test_wampcra_challenge(self, session):
        challenge = Challenge('wampcra', {'challenge': b'\xff\x00345jfsdf'})
        retval = session.onChallenge(challenge)
        assert isinstance(retval, bytes)

    @pytest.mark.parametrize('session', ['wampcra'], indirect=['session'])
    def test_wampcra_salted_challenge(self, session):
        challenge = Challenge('wampcra', {'challenge': b'\xff\x00345jfsdf', 'salt': '5ihod',
                                          'iterations': 5, 'keylen': 32})
        retval = session.onChallenge(challenge)
        assert isinstance(retval, bytes)


class TestWAMPClient:
    @pytest.fixture
    def otherclient(self, request, event_loop, context):
        kwargs = getattr(request, 'param', {})
        kwargs.setdefault('host', os.getenv('CROSSBAR_HOST', 'localhost'))
        kwargs.setdefault('max_reconnection_attempts', 0)
        client = WAMPClient(**kwargs)
        event_loop.run_until_complete(client.start(context))
        yield client
        event_loop.run_until_complete(client.stop())

    @pytest.mark.asyncio
    async def test_client_events(self, wampclient: WAMPClient):
        def listener(event):
            events.append(event)

        events = []
        wampclient.realm_joined.connect(listener)
        wampclient.realm_left.connect(listener)
        await wampclient.connect()
        await wampclient.stop()

        assert len(events) == 2
        assert isinstance(events[0], SessionJoinEvent)
        assert isinstance(events[1], SessionLeaveEvent)

    @pytest.mark.parametrize('connect_first', [False, True])
    @pytest.mark.asyncio
    async def test_call(self, wampclient: WAMPClient, connect_first):
        if connect_first:
            await wampclient.connect()

        result = await wampclient.call('wamp.session.count')
        assert result == 1

    @pytest.mark.asyncio
    async def test_register_call_progress(self, wampclient: WAMPClient):
        async def progressive_procedure(ctx, start, end):
            for value in range(start, end):
                ctx.progress(value)
            return end

        progress_values = []
        await wampclient.register(progressive_procedure, 'test.progressive')
        result = await wampclient.call('test.progressive', 2, 6,
                                       options=CallOptions(on_progress=progress_values.append))
        assert progress_values == [2, 3, 4, 5]
        assert result == 6

    @pytest.mark.asyncio
    async def test_register_call_blocking(self, wampclient: WAMPClient):
        @executor
        def add(ctx, x, y):
            return x + y

        await wampclient.register(add, 'test.add')
        result = await wampclient.call('test.add', 2, 3)
        assert result == 5

    @pytest.mark.asyncio
    async def test_register_call_plain(self, wampclient: WAMPClient):
        def add(ctx, x, y):
            return x + y

        await wampclient.register(add, 'test.add')
        result = await wampclient.call('test.add', 2, 3)
        assert result == 5

    @pytest.mark.parametrize('wampclient', [
        {'auth_method': 'wampcra', 'auth_id': 'testuser', 'auth_secret': 'testpass'}
    ], indirect=True)
    @pytest.mark.asyncio
    async def test_auth_wampcra(self, wampclient: WAMPClient):
        await wampclient.connect()
        result = await wampclient.call('wamp.session.get', wampclient.session_id)
        assert result['authid'] == wampclient.details.authid == 'testuser'

    @pytest.mark.parametrize('wampclient', [
        {'auth_method': 'ticket', 'auth_id': 'device1', 'auth_secret': 'abc123'}
    ], indirect=True)
    @pytest.mark.asyncio
    async def test_auth_ticket(self, wampclient: WAMPClient):
        await wampclient.connect()
        result = await wampclient.call('wamp.session.get', wampclient.session_id)
        assert result['authid'] == wampclient.details.authid == 'device1'

    @pytest.mark.parametrize('wampclient', [
        {'auth_method': 'ticket', 'auth_id': 'device1', 'auth_secret': 'abc124'}
    ], indirect=True)
    @pytest.mark.asyncio
    async def test_auth_failure(self, wampclient: WAMPClient):
        with pytest.raises(ConnectionError) as exc:
            await wampclient.connect()

        assert exc.match('ticket in static WAMP-Ticket authentication is invalid')

    @pytest.mark.asyncio
    async def test_publish_autoconnect(self, wampclient: WAMPClient):
        result = await wampclient.publish('test.topic', options=PublishOptions(acknowledge=True))
        assert result

    @pytest.mark.parametrize('connect_first', [False, True])
    @pytest.mark.asyncio
    async def test_publish_subscribe(self, wampclient: WAMPClient, connect_first):
        async def subscriber(ctx, *args):
            await q.put(args)
            raise Exception()

        q = asyncio.Queue()
        if connect_first:
            await wampclient.connect()

        await wampclient.subscribe(subscriber, 'test.topic')
        publication_id = await wampclient.publish(
            'test.topic', 2, 3, options=PublishOptions(exclude_me=False, acknowledge=True))
        assert isinstance(publication_id, int)
        event = await asyncio.wait_for(q.get(), 2)
        assert event == (2, 3)

    @pytest.mark.parametrize('connect_first', [False, True])
    @pytest.mark.asyncio
    async def test_map_exception(self, wampclient: WAMPClient, connect_first):
        class TestException(Exception):
            pass

        async def error(ctx):
            raise TestException

        if connect_first:
            await wampclient.connect()

        wampclient.map_exception(TestException, 'test.exception')
        await wampclient.register(error, 'test.error')
        with pytest.raises(TestException):
            await wampclient.call('test.error')

    @pytest.mark.asyncio
    async def test_connect_procedure_registration_failure(self, wampclient: WAMPClient,
                                                          otherclient: WAMPClient):
        """
        Test that a failure in registering the registry's procedures causes the connection attempt
        to fail.

        """
        await otherclient.register(lambda ctx: None, 'blah')
        with pytest.raises(ApplicationError):
            await wampclient.register(lambda ctx: None, 'blah')

        assert wampclient.session_id is None

    @pytest.mark.parametrize('wampclient', [
        {'port': 8081, 'max_reconnection_attempts': 1, 'reconnect_delay': 0.3}], indirect=True)
    @pytest.mark.asyncio
    async def test_connect_retry(self, wampclient: WAMPClient, caplog):
        """Test that if the client can't connect, it will retry after a delay."""
        with pytest.raises(ConnectionRefusedError):
            await wampclient.connect()

        messages = [record.message for record in caplog.records
                    if record.name == 'asphalt.wamp.client' and
                    record.message.startswith('Connection failed')]
        assert len(messages) == 1
        assert re.fullmatch("Connection failed \(attempt 1\): ConnectionRefusedError\(.+?\); "
                            "reconnecting in 0.3 seconds", messages[0])

    @pytest.mark.asyncio
    async def test_close_wait_handlers(self, event_loop, wampclient: WAMPClient,
                                       otherclient: WAMPClient, caplog):
        """
        Test that WAMPClient.close() waits for any running handler tasks to finish before
        disconnecting from the router.

        """
        async def sleep_subscriber(ctx):
            nonlocal close_task
            close_task = event_loop.create_task(wampclient.stop())
            await asyncio.sleep(0.3)

        async def sleep_sum(ctx, x, y):
            await asyncio.sleep(0.3)
            return x + y

        caplog.set_level(logging.INFO)
        close_task = None
        await wampclient.register(sleep_sum)
        await wampclient.subscribe(sleep_subscriber, 'testtopic')

        await otherclient.publish('testtopic', options=PublishOptions(acknowledge=True))
        result = await otherclient.call('sleep_sum', 1, 2)
        assert result == 3
        await close_task

        messages = [record.message for record in caplog.records
                    if record.name == 'asphalt.wamp.client' and
                    record.message.startswith('Waiting for')]
        assert messages == ['Waiting for 2 WAMP subscription/procedure handler tasks to finish']

    @pytest.mark.asyncio
    async def test_connect_twice(self, wampclient: WAMPClient):
        """
        Test that when connect() is called while connected, it just returns a Future that resolves
        immediately.

        """
        retval = wampclient.connect()
        assert isinstance(retval, asyncio.Task)
        await retval

        retval = wampclient.connect()
        assert isinstance(retval, asyncio.Future)
        await retval

    def test_session_id_not_connected(self, wampclient: WAMPClient):
        assert wampclient.session_id is None

    def test_session_details_not_connected(self, wampclient: WAMPClient):
        assert wampclient.details is None

    @pytest.mark.parametrize('wampclient', [
        {'auth_method': 'ticket', 'auth_id': 'device1', 'auth_secret': 'abc123'}
    ], indirect=True)
    @pytest.mark.asyncio
    async def test_sentry_extras_provider_procedure(self, wampclient: WAMPClient,
                                                    context: Context, monkeypatch):
        class DummyReporter(ExceptionReporter):
            def report_exception(self, ctx: Context, exception: BaseException, message: str,
                                 extra: Dict[str, Any]) -> None:
                errors.append((exception, message, extra))

        def handler(ctx):
            raise Exception('foo')

        errors = []
        context.add_resource(DummyReporter(), types=[ExceptionReporter])
        context.add_resource(WAMPExtrasProvider(), types=[ExtrasProvider])
        await wampclient.register(handler, 'dummyprocedure')
        monkeypatch.setattr('asphalt.wamp.extras_providers.SENTRY_CLASS_NAME',
                            qualified_name(DummyReporter))
        with pytest.raises(ApplicationError):
            await wampclient.call('dummyprocedure')

        assert len(errors) == 1
        exc, message, extra = errors[0]
        assert type(exc) is Exception
        assert str(exc) == 'foo'
        assert message == "Error running handler for procedure 'dummyprocedure'"
        assert extra == {'extra': {'procedure': 'dummyprocedure'},
                         'user_context': {'auth_role': 'authorized_users',
                                          'id': 'device1',
                                          'session_id': wampclient.session_id}
                         }

    @pytest.mark.parametrize('wampclient', [
        {'auth_method': 'ticket', 'auth_id': 'device1', 'auth_secret': 'abc123'}
    ], indirect=True)
    @pytest.mark.asyncio
    async def test_sentry_extras_provider_subscriber(self, wampclient: WAMPClient,
                                                     context: Context, monkeypatch):
        class DummyReporter(ExceptionReporter):
            def report_exception(self, ctx: Context, exception: BaseException, message: str,
                                 extra: Dict[str, Any]) -> None:
                errors.append((exception, message, extra))

        def handler(ctx):
            ctx.loop.call_soon(event.set)
            raise Exception('foo')

        event = asyncio.Event()
        errors = []
        context.add_resource(DummyReporter(), types=[ExceptionReporter])
        context.add_resource(WAMPExtrasProvider(), types=[ExtrasProvider])
        await wampclient.subscribe(handler, 'dummytopic')
        monkeypatch.setattr('asphalt.wamp.extras_providers.SENTRY_CLASS_NAME',
                            qualified_name(DummyReporter))
        await wampclient.publish('dummytopic', options=dict(acknowledge=True, exclude_me=False))

        await event.wait()
        assert len(errors) == 1
        exc, message, extra = errors[0]
        assert type(exc) is Exception
        assert str(exc) == 'foo'
        assert message == "Error running subscription handler for topic 'dummytopic'"
        assert extra == {'extra': {'topic': 'dummytopic'},
                         'user_context': {'auth_role': 'authorized_users',
                                          'id': 'device1',
                                          'session_id': wampclient.session_id}
                         }



"""
This example demonstrates how to find out the caller's authentication ID (username).
It should be noted that for this to work with Crossbar at least, caller disclosure must be enabled
in the authentication configuration.
"""

import logging

from asphalt.core import ContainerComponent, Context, run_application
from asphalt.wamp.context import CallContext


def hello(ctx: CallContext):
    return 'Hello, {}!'.format(ctx.caller_auth_id)


class RPCServerComponent(ContainerComponent):
    async def start(self, ctx: Context):
        self.add_component('wamp', auth_method='ticket',
                           auth_id='testserver', auth_secret='server123')
        await super().start(ctx)

        await ctx.wamp.register(hello, 'hello')


run_application(RPCServerComponent(), logging=logging.INFO)



import logging
from typing import Any, Dict

import txaio
from asphalt.core import Component, Context, merge_config, context_teardown
from async_generator import yield_
from typeguard import check_argument_types

from asphalt.exceptions import ExtrasProvider
from asphalt.wamp.client import WAMPClient
from asphalt.wamp.extras_providers import WAMPExtrasProvider

__all__ = ('WAMPComponent',)

logger = logging.getLogger(__name__)


class WAMPComponent(Component):
    """Creates one or more :class:`~asphalt.wamp.client.WAMPClient` resources."""

    def __init__(self, clients: Dict[str, Dict[str, Any]] = None, **default_client_args):
        """
        If the ``clients`` argument is omitted or empty, a default client with the context
        attribute ``wamp`` will be created.

        If ``clients`` is defined, any keyword arguments to the component become default options
        for the clients.

        If you wish to change the context attribute of a WAMP client, use the ``context_attr``
        argument.

        :param clients: a dictionary of resource name ⭢ :class:`.WAMPClient` constructor arguments
        :param default_client_args: :class:`.WAMPClient` base options for all clients or arguments
            for the default client if ``clients`` is not specified

        """
        assert check_argument_types()
        if not clients:
            default_client_args.setdefault('context_attr', 'wamp')
            clients = {'default': default_client_args}

        self.clients = []
        for resource_name, config in clients.items():
            config = merge_config(default_client_args, config)
            context_attr = config.pop('context_attr', resource_name)
            client = WAMPClient(**config)
            self.clients.append((resource_name, context_attr, client))

    @context_teardown
    async def start(self, ctx: Context):
        # Autobahn uses txaio to bridge the API gap between asyncio and Twisted so we need to set
        # it up for asyncio here
        txaio.use_asyncio()
        txaio.config.loop = ctx.loop

        ctx.add_resource(WAMPExtrasProvider(), 'wamp', types=[ExtrasProvider])

        for resource_name, context_attr, client in self.clients:
            await client.start(ctx)
            ctx.add_resource(client, resource_name, context_attr)
            logger.info('Configured WAMP client (%s / ctx.%s; host=%s; port=%d; realm=%s)',
                        resource_name, context_attr, client.host, client.port, client.realm)

        await yield_()

        for resource_name, context_attr, client in self.clients:
            await client.stop()
            logger.info('Shut down WAMP client (%s)', resource_name)



"""
This example demonstrates the use of a WAMP registry to register a procedure handler and an event
subscriber.
"""

import logging

from asphalt.core import ContainerComponent, Context, run_application
from asphalt.wamp.context import CallContext, EventContext
from asphalt.wamp.registry import WAMPRegistry

topics = WAMPRegistry('topics')
procedures = WAMPRegistry('test')


@procedures.procedure
async def reverse(ctx: CallContext, message: str):
    return message[::-1]


@topics.subscriber('subtopic')
async def subtopic(ctx: EventContext, message):
    print('message received from topic.subtopic: %s' % message)


class RPCServerComponent(ContainerComponent):
    async def start(self, ctx: Context):
        registry = WAMPRegistry()
        registry.add_from(topics)
        registry.add_from(procedures)
        self.add_component('wamp', registry=registry)
        await super().start(ctx)
        await ctx.wamp.connect()


run_application(RPCServerComponent(), logging=logging.DEBUG)



from typing import Optional, Dict, Any

from asphalt.core import qualified_name, Context
from asphalt.exceptions.api import ExceptionReporter, ExtrasProvider

SENTRY_CLASS_NAME = 'asphalt.exceptions.reporters.sentry.SentryExceptionReporter'


class WAMPExtrasProvider(ExtrasProvider):
    def get_extras(self, ctx: Context, reporter: ExceptionReporter) -> Optional[Dict[str, Any]]:
        from asphalt.wamp import CallContext, EventContext

        extra = None
        if qualified_name(reporter) == SENTRY_CLASS_NAME:
            if isinstance(ctx, CallContext):
                extra = {'extra': {'procedure': ctx.procedure}}
                if ctx.caller_auth_id:
                    extra['user_context'] = {
                        'id': ctx.caller_auth_id,
                        'auth_role': ctx.caller_auth_role,
                        'session_id': ctx.caller_session_id
                    }
            elif isinstance(ctx, EventContext):
                extra = {'extra': {'topic': ctx.topic}}
                if ctx.publisher_auth_id:
                    extra['user_context'] = {
                        'id': ctx.publisher_auth_id,
                        'auth_role': ctx.publisher_auth_role,
                        'session_id': ctx.publisher_session_id
                    }

        return extra



from asphalt.serialization.api import Serializer
from autobahn.wamp import serializer as autobahn
from autobahn.wamp.interfaces import IObjectSerializer


class ObjectSerializerWrapper(IObjectSerializer):
    def __init__(self, serializer: Serializer, binary: bool):
        self._serializer = serializer
        self._binary = binary

    @property
    def BINARY(self) -> bool:
        return self._binary

    def serialize(self, obj) -> bytes:
        return self._serializer.serialize(obj)

    def unserialize(self, payload: bytes):
        return [self._serializer.deserialize(payload)]


class SerializerWrapper(autobahn.Serializer):
    def __init__(self, serializer: Serializer, serializer_id: str, rawsocket_id: int,
                 binary: bool):
        super(SerializerWrapper, self).__init__(ObjectSerializerWrapper(serializer, binary))
        self.SERIALIZER_ID = serializer_id
        self.RAWSOCKET_SERIALIZER_ID = rawsocket_id
        self.MIME_TYPE = serializer.mimetype


def wrap_serializer(serializer: Serializer):
    if serializer.mimetype == 'application/json':
        return SerializerWrapper(serializer, 'json', 1, False)
    elif serializer.mimetype == 'application/msgpack':
        return SerializerWrapper(serializer, 'msgpack', 2, True)
    elif serializer.mimetype == 'application/cbor':
        return SerializerWrapper(serializer, 'cbor', 3, True)

    raise TypeError('cannot adapt serializer of type {}'.format(serializer.mimetype))
